;;;-*-Mode:LISP; Package:(PCL LISP 1000); Base:10; Syntax:Common-lisp -*-
;;;

;;; This version hacked by KDF to allow systems to be stored on RT's and other
;;; unix boxes while being compiled on Symbolics machines 

;;; *************************************************************************
;;; Copyright (c) 1985, 1986, 1987, 1988 Xerox Corporation.
;;; All rights reserved.
;;;
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted.  Any distribution of this
;;; software or derivative works must comply with all applicable United
;;; States export control laws.
;;; 
;;; This software is made available AS IS, and Xerox Corporation makes no
;;; warranty about the software, its performance or its conformity to any
;;; specification.
;;; 
;;; Any person obtaining a copy of this software is requested to send their
;;; name and post office or electronic mail address to:
;;;   CommonLoops Coordinator
;;;   Xerox Artifical Intelligence Systems
;;;   2400 Hanover St.
;;;   Palo Alto, CA 94303
;;; (or send Arpanet mail to CommonLoops-Coordinator.pa@Xerox.arpa)
;;;
;;; Suggestions, comments and requests for improvements are also welcome.
;;; *************************************************************************
;;;
;;; Some support stuff for compiling and loading PCL.  It would be nice if
;;; there was some portable make-system we could all agree to share for a
;;; while.  At least until people really get databases and stuff.
;;;
;;; ***                                                               ***
;;; ***        DIRECTIONS FOR INSTALLING PCL AT YOUR SITE             ***
;;; ***                                                               ***
;;;
;;; To get PCL working at your site you should:
;;; 
;;;  - Get all the PCL source files from Xerox.  The complete list of source
;;;    file names can be found in the defsystem for PCL which appears towards
;;;    the end of this file.
;;; 
;;;  - Edit the variable *pcl-directory* below to specify the directory at
;;;    your site where the pcl sources and binaries will be.  This variable
;;;    can be found by searching from this point for the string "***" in
;;;    this file.
;;; 
;;;  - Use the function (pcl::compile-pcl) to compile PCL for your site.
;;; 
;;;  - Once PCL has been compiled it can be loaded with (pcl::load-pcl).
;;;    Note that PCL cannot be loaded on top of itself, nor can it be
;;;    loaded into the same world it was compiled in.
;;;

(in-package 'pcl :use (list (or (find-package 'walker)
                                (make-package 'walker :use '(lisp)))
                            'lisp))


(defvar *pcl-system-date* "2/2/88  Febuary 2nd 1988")


;;;
;;; Various hacks to get people's *features* into better shape.
;;; 
(eval-when (compile load eval)

  #+Symbolics
  (si:inhibit-style-warnings 
    (let ((major (si:get-release-version))) ;; At least, works for me!
      (cond ((= major 6) (pushnew ':symbolics-release-6 *features*))
            ((= major 7) (pushnew ':symbolics-release-7 *features*))
            (t (error "don't know this system version")))))
  
  (dolist (feature *features*)
    (when (and (symbolp feature)                ;3600!!
               (equal (symbol-name feature) "CMU"))
      (pushnew :CMU *features*)))
  
  #+TI
  (if (eq (si:local-binary-file-type) :xld)
      (pushnew ':ti-release-3 *features*)
      (pushnew ':ti-release-2 *features*))

  #+IBCL
  (push 'kcl *features*)
  
  )



;;; Yet Another Sort Of General System Facility and friends.
;;;
;;; The entry points are defsystem and operate-on-system.  defsystem is used
;;; to define a new system and the files with their load/compile constraints.
;;; Operate-on-system is used to operate on a system defined that has been
;;; defined by defsystem.  For example:
#||

(defsystem my-very-own-system
	   "/usr/myname/lisp/"
  ((classes   (precom)           ()                ())
   (methods   (precom classes)   (classes)         ())
   (precom    ()                 (classes methods) (classes methods))))

This defsystem should be read as follows:

* Define a system named MY-VERY-OWN-SYSTEM, the sources and binaries
  should be in the directory "/usr/me/lisp/".  There are three files
  in the system, there are named classes, methods and precom.  (The
  extension the filenames have depends on the lisp you are running in.)
  
* For the first file, classes, the (precom) in the line means that
  the file precom should be loaded before this file is loaded.  The
  first () means that no other files need to be loaded before this
  file is compiled.  The second () means that changes in other files
  don't force this file to be recompiled.

* For the second file, methods, the (precom classes) means that both
  of the files precom and classes must be loaded before this file
  can be loaded.  The (classes) means that the file classes must be
  loaded before this file can be compiled.  The () means that changes
  in other files don't force this file to be recompiled.

* For the third file, precom, the first () means that no other files
  need to be loaded before this file is loaded.  The first use of
  (classes methods)  means that both classes and methods must be
  loaded before this file can be compiled.  The second use of (classes
  methods) mean that whenever either classes or methods changes precom
  must be recompiled.

Then you can compile your system with:

 (operate-on-system 'my-very-own-system :compile)

and load your system with:

 (operate-on-system 'my-very-own-system :load)

||#

;;; 
(defvar *system-directory*)

;;;
;;; *port* is a list of symbols (in the PCL package) which represent the
;;; Common Lisp in which we are now running.  Many of the facilities in
;;; defsys use the value of *port* rather than #+ and #- to conditionalize
;;; the way they work.
;;; 
(defvar *port*
        '(#+Symbolics             Symbolics
          #+Symbolics-Release-6   Rel-6
          #+Symbolics-Release-7   Rel-7
          #+Lucid                 Lucid
          #+Xerox                 Xerox
          #+TI                    TI
          #+(and dec vax common)  Vaxlisp
          #+KCL                   KCL
          #+excl                  excl
          #+:CMU                  CMU
          #+HP                    HP
          #+:gclisp               gclisp
          #+pyramid               pyramid
          #+:coral                coral))

;;;
;;; When you get a copy of PCL (by tape or by FTP), the sources files will
;;; have extensions of ".lisp" in particular, this file will be defsys.lisp.
;;; The preferred way to install pcl is to rename these files to have the
;;; extension which your lisp likes to use for its files.  Alternately, it
;;; is possible not to rename the files.  If the files are not renamed to
;;; the proper convention, the second line of the following defvar should
;;; be changed to:
;;;     (let ((files-renamed-p nil)
;;;
;;; Note: Something people installing PCL on a machine running Unix
;;;       might find useful.  If you want to change the extensions
;;;       of the source files from ".lisp" to ".lsp", *all* you have
;;;       to do is the following:
;;;
;;;       % foreach i (*.lisp)
;;;       ? mv $i $i:r.lsp
;;;       ? end
;;;       %
;;;
;;;       I am sure that a lot of people already know that, and some
;;;       Unix hackers may say, "jeez who doesn't know that".  Those
;;;       same Unix hackers are invited to fix mv so that I can type
;;;       "mv *.lisp *.lsp".
;;;
(defvar *pathname-extensions*
  (let ((files-renamed-p t)
        (proper-extensions
          (car
           '(#+Symbolics                         ("lisp"  . "bin")
             #+(and dec common vax (not ultrix)) ("LSP"   . "FAS")
             #+(and dec common vax ultrix)       ("lsp"   . "fas")
             #+KCL                               ("lsp"   . "o")
             #+Xerox                             ("lisp"  . "dfasl"
	     #+(and Lucid :RT)                    ("lisp"  . "bbin"))
             #+(and Lucid MC68000)               ("lisp"  . "lbin")
             #+(and Lucid VAX VMS)               ("lisp"  . "vbin")
             #+(and Lucid Prime)                 ("lisp"  . "pbin")
	     #+(and Lucid SUNRise)               ("lisp"  . "sbin")
             #+(and excl allegro)                ("lisp"  . "fasl")
	     #+(and excl (not allegro))          ("cl"    . "fasl")
             #+:CMU                              ("slisp" . "sfasl")
             #+HP                                ("l"     . "b")
             #+TI ("lisp" . #.(string (si::local-binary-file-type)))
             #+:gclisp                           ("LSP"   . "F2S")
             #+pyramid                           ("clisp" . "o")
             #+:coral                            ("lisp"  . "fasl")
             ))))
    (cond ((null proper-extensions) '("l" . "lbin"))
          ((null files-renamed-p) (cons "lisp" (cdr proper-extensions)))
          (t proper-extensions))))

(eval-when (compile load eval)

(defun get-system (name)
  (get name 'system-definition))

(defun set-system (name new-value)
  (setf (get name 'system-definition) new-value))

(defmacro defsystem (name directory &body files)
  `(set-system ',name (cons (pathname ,directory) ',files)))

)


;;;
;;; The internal datastructure used when operating on a system.
;;; 
(defstruct (module (:constructor make-module (name))
                   (:print-function
                     (lambda (m s d)
                       (declare (ignore d))
                       (format s "#<Module ~A>" (module-name m)))))
  name
  load-env
  comp-env
  recomp-reasons)

(defun make-modules (system-description)
  (let ((modules ()))
    (labels ((get-module (name)
               (or (find name modules :key #'module-name)
                   (progn (setq modules (cons (make-module name) modules))
                          (car modules))))
             (parse-spec (spec)
               (if (eq spec 't)
                   (reverse (cdr modules))                   
                   (mapcar #'get-module spec))))
      (dolist (file system-description)
        (let* ((name (car file))
               (port (car (cddddr file)))
               (module nil))
          (when (or (null port)
                    (member port *port*))
            (setq module (get-module name))
            (setf (module-load-env module) (parse-spec (cadr file))
                  (module-comp-env module) (parse-spec (caddr file))
                  (module-recomp-reasons module) (parse-spec
                                                   (cadddr file))))))
      (let ((filenames (mapcar #'car system-description)))
	(sort modules #'(lambda (name1 name2)
			  (member name2 (member name1 filenames)))
	      :key #'module-name)))))

(defun make-transformations (modules filter make-transform)
  (let ((transforms (list nil)))
    (dolist (m modules)
      (when (funcall filter m transforms)
        (funcall make-transform m transforms)))
    (reverse (cdr transforms))))

(defun make-compile-transformation (module transforms)
  (unless (dolist (trans transforms)
            (and (eq (car trans) ':compile)
                 (eq (cadr trans) module)
                 (return trans)))    
    (dolist (c (module-comp-env module))
      (make-load-transformation c transforms))
    #+symbolics-release-6 (make-load-transformation module transforms)
    (push `(:compile ,module) (cdr transforms)))
  transforms)

(defun make-load-transformation (module transforms)
  (unless (dolist (trans transforms)
            (when (eq (cadr trans) module)
              (cond ((eq (car trans) ':compile) (return nil))
                    ((eq (car trans) ':load)   (return trans)))))
    (dolist (l (module-load-env module))
      (make-load-transformation l transforms))
    (push `(:load ,module) (cdr transforms)))
  transforms)

(defun make-load-without-dependencies-transformation (module transforms)
  (unless (dolist (trans transforms)
            (and (eq (car trans) ':load)
                 (eq (cadr trans) module)
                 (return trans)))
    (push `(:load ,module) (cdr transforms)))
  transforms)

(defun compile-filter (module transforms)
  (or (dolist (r (module-recomp-reasons module))
        (when (dolist (transform transforms)
                (when (and (eq (car transform) ':compile)
                           (eq (cadr transform) r))
                  (return t)))
          (return t)))
      (null (probe-file (make-binary-pathname (module-name module))))
  ;; Some systems aren't civilized enough to pass on a write date
  ;; in such cases, just recompile.  -- KDF
      (let ((source-written (file-write-date (make-source-pathname (module-name module))))
	    (binary-written (file-write-date (make-binary-pathname (module-name module)))))
	(if (and (numberp source-written)
		 (numberp binary-written))
	    (> source-written binary-written)
	    t))))

(defun operate-on-system (name mode &optional arg print-only)
  (let* ((system (get-system name))
	 (*system-directory* (car system))
	 (files (cadr system)))
    (unless system (error "Can't find system with name ~S." name))
    (let ((modules (make-modules files))
	  (transformations ())
	  (confirmed t))
      (flet ((load-module (m)
	       (let ((name (module-name m))
		     (*load-verbose* nil))
		 (when (or confirmed (y-or-n-p "Load ~A" name))
		   (if (dolist (trans transformations)
			 (and (eq (car trans) :compile)
			      (eq (cadr trans) m)
			      (return trans)))
		       (progn (format t "~&Loading source of ~A..." name)
			      (or print-only
				  (load (make-source-pathname name))))
		       (progn (format t "~&Loading binary of ~A..." name)
			      (or print-only
				  (load (make-binary-pathname name))))))))
	     (compile-module (m)
	       (format t "~&Compiling ~A..." (module-name m))
	       (unless print-only
		 (let  ((name (module-name m)))
		   (compile-file (make-source-pathname name)
				 :output-file
				 (make-binary-pathname name)))))
	     (true (&rest ignore) (declare (ignore ignore)) 't))
	(setq transformations
	      (ecase mode
		(:compile
		  ;; Compile any files that have changed and any other files
		  ;; that require recompilation when another file has been
		  ;; recompiled.
		  (make-transformations
		    modules
		    #'compile-filter
		    #'make-compile-transformation))
		(:recompile
		  ;; Force recompilation of all files.
		  (make-transformations
		    modules
		    #'true
		    #'make-compile-transformation))
		(:recompile-some
		  ;; Force recompilation of some files.  Also compile the
		  ;; files that require recompilation when another file has
		  ;; been recompiled.
		  (make-transformations
		    modules
		    #'(lambda (m transforms)
			(or (member (module-name m) arg)
			    (compile-filter m transforms)))
		    #'make-compile-transformation))
		(:query-compile
		  ;; Ask the user which files to compile.  Compile those
		  ;; and any other files which must be recompiled when
		  ;; another file has been recompiled.
		  (make-transformations
		    modules
		    #'(lambda (m transforms)
			(or (compile-filter m transforms)
			    (y-or-n-p "Compile ~A?"
				      (module-name m))))
		    #'make-compile-transformation))
		(:confirm-compile
		  ;; Offer the user a chance to prevent a file from being
		  ;; recompiled.
		  (make-transformations
		    modules
		    #'(lambda (m transforms)
			(and (compile-filter m transforms)
			     (y-or-n-p "Go ahead and compile ~A?"
				       (module-name m))))
		    #'make-compile-transformation))
		(:load
		  ;; Load the whole system.
		  (make-transformations
		    modules
		    #'true
		    #'make-load-transformation))
		(:query-load
		  ;; Load only those files the user says to load.
		  (make-transformations
		    modules
		    #'(lambda (m transforms)
			(declare (ignore transforms))
			(y-or-n-p "Load ~A?" (module-name m)))
		    #'make-load-without-dependencies-transformation))))
	
	(#+Symbolics compiler:compiler-warnings-context-bind
	 #-Symbolics progn      
	 (loop (when (null transformations) (return t))
	       (let ((transform (pop transformations)))
		 (ecase (car transform)
		   (:compile (compile-module (cadr transform)))
		   (:load (load-module (cadr transform)))))))))))


(defun make-source-pathname (name) (make-pathname-internal name :source))
(defun make-binary-pathname (name) (make-pathname-internal name :binary))

(defun make-pathname-internal (name type)
  (let* ((extension (ecase type
                      (:source (car *pathname-extensions*))
                      (:binary (cdr *pathname-extensions*))))
         (directory (etypecase *system-directory*
                      (pathname *system-directory*)
                      (cons (ecase type
                              (:source (car *system-directory*))
                              (:binary (cdr *system-directory*))))))
         (pathname
           (make-pathname
             :name #-VMS (string-downcase (string name))
                   #+VMS (string-downcase (substitute #\_ #\- (string name)))
             :type extension
             :defaults directory)))

;    #+Symbolics
;    (setq pathname (zl:send pathname :new-raw-name (pathname-name pathname))
;          pathname (zl:send pathname :new-raw-type (pathname-type pathname)))

    pathname))



;;; ***                SITE SPECIFIC PCL DIRECTORY                        ***
;;;
;;; *pcl-directory* is a variable which specifies the directory pcl is stored
;;; in at your site.  If the value of the variable is a single pathname, the
;;; sources and binaries should be stored in that directory.  If the value of
;;; that directory is a cons, the CAR should be the source directory and the
;;; CDR should be the binary directory.
;;; 
;(defvar *pcl-directory*
;  #+Symbolics                (cons (pathname "BD:>pcl>")
;     #+Symbolics-release-6         (pathname "BD:>pcl>rel6>")
;     #+Symbolics-release-7         (pathname "BD:>pcl>"))
;  #+Lucid                    (pathname "/usr/pcl/")
;  #+ExCL                     (pathname "/usr/pcl/")
;  #+KCL                      (pathname "/usr/pcl/")
;  #+(and DEC common vax VMS) (pathname "")
;  #+:CMU                     (pathname "pcl:")
;  #+HP                       (pathname "")
;  #+Xerox                    (pathname "{phylum}<pcl>")
;  #+:gclisp                  (pathname "/pcl/")
;  #+pyramid                  (pathname "")
;  #+:coral                   (pathname "rodime:coral:pcl:")
;  )
;
;(defsystem pcl
;	   *pcl-directory*
;  ;; file         load           compile           files which      port
;  ;;              environment    environment       force the of
;  ;;                                               recompilation
;  ;;                                               of this file
;  ((rel-6-patches t              t                ()               rel-6)
;   (rel-7-patches t              t                ()               rel-7)
;   (ti-patches    t              t                ()               ti)
;   (pyr-patches   t              t                ()               pyramid)
;   
;   (pkg          t               t                ())
;   (walk         (pkg)           (pkg)            ())
;   (macros       (pkg walk)      (pkg walk)       ())
;   (low          (pkg walk)      (pkg macros)     (macros))
;   
;   
;   (3600-low     (low)           (low)            (low)            Symbolics)
;   (lucid-low    (low)           (low)            (low)            Lucid)
;   (Xerox-low    (low)           (low)            (low)            Xerox)
;   (ti-low       (low)           (low)            (low)            TI)
;   (vaxl-low     (low)           (low)            (low)            vaxlisp)
;   (kcl-low      (low)           (low)            (low)            KCL)
;   (excl-low     (low)           (low)            (low)            excl)
;   (cmu-low      (low)           (low)            (low)            CMU)
;   (hp-low       (low)           (low)            (low)            HP)
;   (gold-low     (low)           (low)            (low)            gclisp) 
;   (pyr-low      (low)           (low)            (low)            pyramid) 
;   (coral-low    (low)           (low)            (low)            coral)
;   
;   (fin          t               t                (low))
;   (defs         t               t                (macros))
;   (boot         t               t                (defs fin))
;   (slots        t               t                (boot defs fin))
;   (defclass     t               t                (boot defs fin))
;   (std-class    t               t                (boot defs fin))
;   (braid1       t               t                (boot defs fin))
;   (fsc          t               t                (boot defs fin))
;   (methods      t               t                (boot defs fin))
;   (combin       t               t                (boot defs fin))
;   (dcode        t               t                (defs fin))
;   (dcode-pre1   t               t                (defs fin dcode))
;   (fixup        t               t                (boot defs fin))
;   (high         t               t                (boot defs fin))
;   (compat       t               t                ())
;   
;   (7debug        t              t                ()               rel-7)
;   ))
;
;(defun compile-pcl (&optional m)
;  (cond ((null m)        (operate-on-system 'pcl :compile))
;        ((eq m :print)   (operate-on-system 'pcl :compile () t))
;        ((eq m :query)   (operate-on-system 'pcl :query-compile))
;        ((eq m :confirm) (operate-on-system 'pcl :confirm-compile))
;        ((eq m 't)       (operate-on-system 'pcl :recompile))        
;        ((listp m)       (operate-on-system 'pcl :compile-from m))
;	((symbolp m)     (operate-on-system 'pcl :recompile-some `(,m)))))
;
;(defun load-pcl (&optional m)
;  (cond ((null m)      (operate-on-system 'pcl :load))
;        ((eq m :query) (operate-on-system 'pcl :query-load))))